"""
lotterylib by Jonas 'ojii' Obrist

Compatible with esl.ojii.ch API Protocol Version 1.0

For documentation visit http://esl.ojii.ch/docs/lotterylib-py/
"""
#===============================================================================
# Imports
#===============================================================================
from xml.etree.ElementTree import fromstring
#===============================================================================
# Public API
#===============================================================================
class InvalidProtocol(Exception): pass


class Application(object):
    def __init__(self, api_key, server_key, app_key, protocol=1):
        """
        Set up a new application with a api_key and server_key to get information
        from the web and an app_name to identify it.
        """
        self.api_key = api_key
        self.server_key = server_key
        self.app_key = app_key
        self.protocol = get_protocol(protocol, self)
    
    def __getattr__(self, attr):
        return self.protocol._handle(attr)
    
    
def set_downloader(downloader):
    """
    Set a custom downloader. The downloader should be a subclass of
    BaseDownloader.
    """
    manager.set_downloader(downloader)
    
def get_history(lottery, limit=10, callback=None, async=True, protocol=1):
    """
    Return the last 'limit' results of 'lottery'.
    """
    return get_protocol(protocol).get_history(lottery, limit, callback, async)
    

#===============================================================================
# Downloaders
#===============================================================================

class BaseDownloader(object):
    def __init__(self, url, callback, response_builder):
        self.url = url
        self.callback = callback
        self.response_builder = response_builder
        
    def run(self, async):
        if async:
            return self.handle_async()
        else:
            return self.handle()
        
    def handle_async(self):
        """
        Download self.url asynchronously and call self.callback with the result.
        """
        self.callback(self.handle())
        
    def handle(self):
        """
        Download self.url and return the result.
        """
        return self.response_builder('error:no downloader')


class Urllib2Downloader(BaseDownloader):
    """
    A threaded urllib2 downloader, this is the default fallback on non-unix
    systems or unix systems without wget
    """
    import threading
    import functools
    import urllib2

    def handle_async(self):
        """
        Handle asynchronous calls. We do this by running a threading.Thread with
        the downloader function.
        """
        def downloader():
            try:
                u = self.urllib2.urlopen(self.url)
            except self.urllib2.URLError, e:
                callback(self.response_builder('urrlib2.URLError: %s' % e.message, error=True))
            self.callback(self.response_builder(u.read()))
        self.threading.Thread(target=self.functools.partial(downloader)).start()

    def handle(self):
        """
        Handle synchronous calls. Since a downloader should never raise an
        exception, we fetch URLErrors and instead return a Response object wich
        failed.
        """
        try:
            u = self.urllib2.urlopen(self.url)
        except self.urllib2.URLError:
            callback(self.response_builder(e.message, error=True))
        return self.response_builder(u.read())

    
class WgetDownloader(BaseDownloader):
    """
    Downloader using the wget application. This runs the download in another
    process and thus creates minimal lag. The downloader checks periodically
    using gamethread whether the download finished or not.
    """
    try:
        import gamethread
        _debug = False
    except ImportError:
        _debug = True
    import os
    tmp_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'lotterylib_tmp_file')
    
    def __init__(self, *args, **kwargs):
        BaseDownloader.__init__(self, *args, **kwargs)
        if self.os.path.exists(self.tmp_file):
            self.os.remove(self.tmp_file)
        self.timer = 0
        
    def handle_async(self):
        self.os.system('wget -q --output-document=%s %s' % (self.tmp_file, self.url))
        self.schedule()
        
    def schedule(self):
        if self._debug:
            import time
            time.sleep(2)
            self.check()
        else:
            self.gamethread.delayed(2, self.check)
        
    def check(self):
        if self.os.path.exists(self.tmp_file):
            f = open(self.tmp_file,'rb')
            c = f.read()
            f.close()
            self.callback(self.response_builder(c))
        else:
            self.timer += 1
            if self.timer < 30:
                self.schedule()
            else:
                self.callback(self.response_builder('connection timeout', error=True))
                
    def handle(self):
        pipe = self.os.popen('wget -q -O - %s' % self.url)
        response = pipe.read()
        return self.response_builder(response)
    
    
#===============================================================================
# Protocol
#===============================================================================
PROTOCOL_HANDLERS = {}
class BaseProtocol(object):
    VERSION = '0'
    BASE_URL = 'http://esl.ojii.ch/api/%s/%s/%s'
    
    def __init__(self, application=None):
        self.app = application
    
    def build_url(self, command, *args):
        argstring = '/'.join(map(str, args))
        return self.BASE_URL % (self.VERSION, command, argstring)
    
    def _handle(self, attr):
        return getattr(self, attr)

    def __getattr__(self, attr):
        raise NotImplementedError, "Method %s is not implemented in Protocol Version %s" % (attr, self.VERSION)


class ProtocolV1(BaseProtocol):
    """
    Version 1 of the API Protocol
    """
    class ResponseBuilder:
        """
        Just a namespace for all response builders
        """
        @staticmethod
        def winners(content, error=False):
            if error:
                return ResponseV1(errmsg=content)
            if not content:
                return ResponseV1(errmsg='No Response')
            lines = content.split('\n')
            header = lines.pop(0)
            if not header:
                return ResponseV1(errmsg='No Header')
            if header.startswith('error:'):
                return ResponseV1(errmsg=header[6:])
            if not header == 'ok':
                return ResponseV1(errmsg='Unexpected Response: %s' % header)
            result = {}
            current = None
            for line in lines:
                if line.startswith('@'):
                    current = line[1:]
                    result[current] = []
                elif current:
                    result[current].append(line)
            return ResponseV1(result=result)
        
        @staticmethod
        def history(content, error=False):
            if error:
                return ResponseV1(errmsg=content)
            try:
                return ResponseV1(django_xml_to_python(content))
            except Exception, e:
                return ResponseV1(None, '%s: %s' % (e.__class__.__name__, e.message))
                
    # This is version 1
    VERSION = '1'
    
    def _generic(self, callback, async, response_builder, command, *args):
        """
        Generic protocol handler
        """
        downloader = get_downloader()
        url = self.build_url(command, *args)
        if async:
            if not callable(callback):
                raise TypeError, "'callback' argument for '%s' must be callable if asynchronous flag is True" % command
            dl = downloader(url, callback, response_builder)
        else:
            dl = downloader(url, callback, response_builder)
        return dl.run(async)
    
    def get_queue(self, callback=None, async=True):
        """
        Calls 'callback' with a list of steamid's that won since the last check
        """
        return self._generic(callback, async, self.ResponseBuilder.winners, 'get_queue', self.app.api_key, self.app.server_key, self.app.app_key)
    
    def get_history(self, lottery, limit=10, callback=None, async=True):
        """
        Get the history for a lottery.
        """
        return self._generic(callback, async, self.ResponseBuilder.history, 'get_history', lottery, 'xml', limit)
PROTOCOL_HANDLERS['1'] = ProtocolV1
    
def get_protocol(version, application=None):
    version = str(version)
    return PROTOCOL_HANDLERS.get(version, BaseProtocol)(application)

#===============================================================================
# Response
#===============================================================================

class ResponseError(Exception):
    """
    Error which occurs if success-only methods are used in an unsuccessful 
    Response object
    """


class Response(object):
    """
    Since lotterylib encourages usage of asynchronous downloads, we can't just
    raise Exceptions. Thus any download (asynchronous or synchronous) will get a
    Response object.
    """
    def __init__(self, result={}, errmsg=''):
        self.success = not errmsg
        self.error = bool(errmsg)
        self.result = result
        self.errmsg = errmsg
        
    def __iter__(self):
        if self.error:
            self.raise_error()
        else:
            return iter(self.result)
        
    def __bool__(self):
        return self.success
        
    def raise_error(self):
        raise ResponseError, self.errmsg
    
class ResponseV1(Response):
    """
    Response objects for version 1 of the API protocol.
    """

#===============================================================================
# Internal
#===============================================================================

def django_xml_to_python(xml):
    et = fromstring(xml)
    objects = []
    for obj in et.getchildren():
        objdict = {}
        for child in obj.getchildren():
            objdict[child.get('name')] = child.text
        objects.append(objdict)
    return objects

def get_default_downloader():
    """
    On Linux it returns the wget_downloader if wget is installed, otherwise it
    returns the urllib2_downloader
    """
    import os
    if os.name == 'posix':
        if os.popen('wget -V').read().lower().startswith('gnu wget'):
            return WgetDownloader
    return Urllib2Downloader

class Manager(object):
    def __init__(self):
        self.downloader = None
        
    def get_downloader(self):
        if self.downloader:
            return self.downloader
        self.downloader = get_default_downloader()
        return self.downloader
    
    def set_downloader(self, downloader):
        self.downloader = downloader
        
manager = Manager()
get_downloader = manager.get_downloader